/**
*  the WS is a websocket abstract class
*/




/**
* module dependencies
*/
const DEBUG = 'ping-ws-client'; 
const debug = require('debug')(DEBUG);
const assert = require('assert');
const EventEmitter = require('events');
const parseOptions = require('./options');
const rowraw = require('ddv-rowraw');
const is = require('is-type-of');
const api = require('ddv-restful-api');
const session = require('ddv-restful-api/lib/session');
const axios = require('axios');


const nodeWs = typeof WebSocket === 'undefined' ? require('ws'):null;



class WS extends EventEmitter {
  /**
   * @constructor
   * @param {Object} options
   */
  constructor(options) {
    super();
    debug('new a WS instance for option: %j', options);
    assert(!is.nullOrUndefined(options), 'should define options object!');
    assert(!is.nullOrUndefined(options.wsUrl), 'should define ws url!');
    assert(!is.nullOrUndefined(options.apiHost), 'should define apiHost!');
    assert(!is.nullOrUndefined(options.uid), 'should define uid!');
    if (is.nullOrUndefined(options) || !options.wsUrl) {
      throw new Error('should define options.wsUrl!');
    }
    if (is.nullOrUndefined(options) || !options.apiHost) {
      throw new Error('should define options.apiHost!');
    }
    if (is.nullOrUndefined(options) || !options.uid) {
      throw new Error('should define options.uid!');
    }

    this.reTryNumSum = 0;

    // parse the options
    this.options = parseOptions(options);
    debug('this options:%j', this.options);
    this.ws = null;
    this.wsUrl = this.options.wsUrl ? this.options.wsUrl : this.wsUrl;

    //设置默认请求域名
    api.setBaseUrl(this.options.apiHost);
    //自定义头前缀
    api.setHeadersPrefix('x-hz-');
    //是否长期会话
    session.setLongStorage(false);
    //设置会话初始化最大自动尝试次数，默认3次
    session.setSessionInitTrySum(3);
    //设置初始化session的path，默认/session/init
    session.setSessionInitPath('/session/init');

  }
  get CONNECTING() { return WS.CONNECTING; }
  get CLOSING() { return WS.CLOSING; }
  get CLOSED() { return WS.CLOSED; }
  get OPEN() { return WS.OPEN; }

  /**
   * check current env
   * @method isBroswer
   * @return {Boolean} 检测当前是否是浏览器环境
   */
  isBroswer() {
    return typeof Window !== 'undefined' && typeof WebSocket !== 'undefined';
  }
  /**
   * send message
   * @param {Object} message
   * @param {Function} callBack
   */
  send(message,  callBack) {
    if (this.readyState === WS.CLOSED || is.nullOrUndefined(this.ws)) {
      this.open();
    }
    debug('ws send message : %j', message);
    assert(!is.nullOrUndefined(message), `should define message object!`);
    if (this.connected && message && this && this.ws && this.ws.send && is.function(this.ws.send)) {
      var _that = this;
      return new Promise(function (resolve, reject) {
        try {
          _that.ws.send(message);
          resolve();
        } catch (e) {
          reject(e);
        }
      }).then(() => {
        if (callBack && is.function(callBack)) {
          callBack();
        }
      });
    }
  }

  /**
   * open the ws connection
   *
   */
  open() {
    debug('ws open wsUrl: %s', this.wsUrl);
    if (this.readyState === WS.OPEN || !this.wsUrl || this.readyState === WS.CONNECTING)
      return;
    this.readyState = WS.CONNECTING;
    this.ws = this.isBroswer() ? new WebSocket(this.wsUrl) : new nodeWs(this.wsUrl);
    this.initBingding();
    this.initEvents();
  }
  /**
   * close the ws connection
   */
  close() {
    if (this && this.ws && this.ws.close) {
      this.ws.close();
      this.connected = false;
      this.readyState = WS.CLOSED;
    }
  }


  /**
   * auto reconnection
   */
  autoReconnection() {
    debug('autoReconnection:%s , reTryNum:%s', new Date(),this.reTryNumSum);

    if(this.reTryNumSum && this.reTryNumSum > this.options.reTryNum){
      debug('more than the max reTryNum:%s',this.options.reTryNum);
      return;
    }

    // 尝试次数
    this.reTryNumSum = !this.reTryNumSum ? 1 : this.reTryNumSum+1;

    if (this.readyState !== WS.CLOSED)
      return;
    if (this.options.isAutoReconnection) {
      this.timer = setTimeout(function () {
        clearTimeout(this.timer);
        this.open();
      }.bind(this), this.options.connectionTime);
    }
  }


  /**
   *
   * on ws connection error emit
   *
   * @param {Object} e
   */
  onError(e) {
    debug('ws on %s errored, error desc: %j !', new Date(), e);
    this.disposeWS();
    this.autoReconnection();
  }

  /**
   * dispose the ws object
   */
  disposeWS() {
    debug(' dispose this ws object : %s', new Date());
    this.connected = false;
    this.readyState = WS.CLOSED;
    this.ws && this.ws.removeAllListeners && this.ws.removeAllListeners() && delete this.ws;
    this.removeAllListeners && this.removeAllListeners();
  }

  /**
   *
   * on ws connection opened emit
   *
   */
  onOpen() {
    debug('ws on %s open the connection', new Date());
  }

  /**
   *
   * on ws connection closed emit
   *
   * @param {Object} e
   */
  onClose(e) {
    debug('ws on %s closed, close desc: %j!', new Date(), e);

    this.disposeWS();
    this.autoReconnection();
  }


  /**
   *  on ws receive message emit
   *
   * @param {Object} message
   */
  onMessage(message) {
    debug('on %s receive message: %j', new Date(), message);

    let msg = JSON.parse(message);

    //如果是首次连接会返回连接ID需要调一个bind接口
    if (msg && msg.type && msg.type === 'connection' && msg.client_id) {

      api.get(`${this.options.apiHost}/v4_0/api/bind?client_id=${msg.client_id}&uid=${this.options.uid}`)
      .send({}).then((res)=>{
        debug('request the bind api success res: %j', res);
        this.readyState = WS.OPEN;
        this.connected = true;
        this.emit('binded');
      }).catch((error) => {
        debug('request the bind api errored res: %j', error);
        this.readyState = WS.CLOSED;
        this.connected = false;
      });
      
      /*axios.get(`${this.options.apiHost}/v4_0/api/bind?client_id=${msg.client_id}&uid=${this.options.uid}`)
        .then((res) => {
          debug('request the bind api success res: %j', res);
          this.readyState = WS.OPEN;
          this.connected = true;
        })
        .catch((error) => {
          debug('request the bind api errored res: %j', error);
          this.readyState = WS.CLOSED;
          this.connected = false;
        });*/
    }
  }
  /**
   * init ws events
   */
  initEvents() {
    this.on('open', this.onOpen.bind(this));
    this.on('error', this.onError.bind(this));
    this.on('close', this.onClose.bind(this));
    this.on('message', this.onMessage.bind(this));
  }
  /**
   * init ws events bingding to this events
   */
  initBingding() {
    // broswer
    if (this.isBroswer()) {
      this.ws.onopen = function () {
        this.emit('open');
      }.bind(this);
      this.ws.onclose = function (e) {
        this.emit('close', e);
      }.bind(this);
      this.ws.onerror = function (e) {
        this.emit('error', e);
      }.bind(this);
      this.ws.onmessage = function (e) {
        this.emit('message', (e.data || e));
      }.bind(this);

    } else { // node
      this.ws.on('open', function onopen() {
        this.emit('open');
      }.bind(this));
      this.ws.on('close', function onclose(e) {
        this.emit('close', e);
      }.bind(this));
      this.ws.on('error', function onerror(e) {
        this.emit('error', e);
      }.bind(this));
      this.ws.on('message', function onmessage(data) {
        this.emit('message', data);
      }.bind(this));
    }
  }
}
/**
 * 连接中
 */
WS.CONNECTING = 0;
/**
 * 连接已打开
 */
WS.OPEN = 1;
/**
 * 关闭中
 */
WS.CLOSING = 2;
/**
 * 已关闭
 */
WS.CLOSED = 3;

module.exports = WS;
